// Copyright (C) 1999 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of APE.
// 
// The exception is that, if you link the APE library with other files
// to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the APE library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name APE.  If you copy code from other releases into a copy of
// APE, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for APE, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#ifndef	__APE_FILE_H__
#define	__APE_FILE_H__

#ifndef	__APE_THREAD_H__
#include <APE/thread.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>

typedef	unsigned long pos_t;

enum
{
	FILE_OPEN_READONLY = O_RDONLY,
	FILE_OPEN_WRITEONLY = O_WRONLY,
	FILE_OPEN_READWRITE = O_RDWR,
	FILE_OPEN_APPEND = O_WRONLY | O_APPEND,

#ifdef	O_SYNC
	FILE_OPEN_SYNC = O_RDWR | O_SYNC,
#else
	FILE_OPEN_SYNC = O_RDWR,
#endif
	FILE_OPEN_TRUNCATE = O_RDWR | O_TRUNC
};

enum
{
	FILE_MAPPED_READ = O_RDONLY,
	FILE_MAPPED_WRITE = O_WRONLY,
	FILE_MAPPED_RDWR = O_RDWR
};

enum
{
	FILE_PERM_PRIVATE = S_IRUSR | S_IWUSR,
	FILE_PERM_GROUP = FILE_PERM_PRIVATE | S_IRGRP | S_IWGRP,
	FILE_PERM_PUBLIC = FILE_PERM_GROUP | S_IROTH | S_IWOTH
};

class File
{
protected:
	int _fd;

	virtual void Initial(const char *fname, int perm)
		{return;};

public:
	File(const char *fname, int access);
	File(const char *fname, int access, int perm);
	File(const File &f);
	File(int fd);

	virtual ~File()
		{close(_fd);};

	int Read(void *buf, size_t len)
		{return read(_fd, (char *)buf, len);};

	int Write(void *buf, size_t len)
		{return write(_fd, (char *)buf, len);};

	int Write(char *buf)
		{return write(_fd, buf, strlen(buf));};

	pos_t Position(pos_t pos)
		{return (pos_t)lseek(_fd, pos, SEEK_SET);}; 

	pos_t Append(void)
		{return (pos_t)lseek(_fd, 0l, SEEK_END);}; 

	File &operator=(const File &f);

	friend read(File &f, void *buf, size_t len)
		{return read(f._fd, (char *)buf, len);};

	friend write(File &f, void *buf, size_t len)
		{return write(f._fd, (char *)buf, len);};

	friend write(File &f, char *buf)
		{return write(f._fd, buf, strlen(buf));};

	friend pos_t append(File &f)
		{return (pos_t)lseek(f._fd, 0l, SEEK_END);};

	friend pos_t position(File &f, pos_t pos)
		{return (pos_t)lseek(f._fd, pos, SEEK_SET);};
};

class NewFile : public File
{
protected:
	void Initial(char *fname, int perm)
		{remove(fname);};

public:
	NewFile(const char *fname, int access, int perms);
};

class NamedPipe : public File
{
protected:
	void Initial(char *fname, int perm);

public:
	NamedPipe(const char *fname, int perms);
};	

class MappedFile : protected File
{
private:
	size_t _len;
	char *_mem;

public:
	MappedFile(const char *fname, pos_t offset, size_t size, int mode);	
	virtual ~MappedFile();

	inline void Sync(void)
		{msync(_mem, _len, MS_SYNC);};

	inline void Update(unsigned offset, size_t len)
		{msync(_mem + offset, len, MS_ASYNC);};

	inline void *Fetch(unsigned offset)
		{return _mem + offset;};

	friend inline void sync(MappedFile &m)
		{msync(m._mem, m._len, MS_SYNC);};

	friend inline void update(MappedFile &m, unsigned offset, size_t len)
		{msync(m._mem + offset, len, MS_ASYNC);};

	friend inline void *fetch(MappedFile &m, unsigned offset = 0)
		{return m._mem + offset;};
};

class LockedFile : protected File
{
public:
	LockedFile(const char *fname, int access);
	LockedFile(const LockedFile &f);

	~LockedFile()
		{close(_fd);};

	int Append(void *buf, size_t len);
	int Request(pos_t pos, void *buf, size_t len);
	int Update(pos_t pos, void *buf, size_t len);
	int Clear(pos_t pos, size_t len);
};

class SharedFile : public File, public Mutex
{
public:
	SharedFile(const char *fname, int access, int perm);
	SharedFile(const char *fname, int access);
	SharedFile(const SharedFile &f);
	SharedFile(int fd);

	~SharedFile()
		{close(_fd);};

	int Append(void *buf, size_t len);
	int Read(pos_t pos, void *buf, size_t len);
	int Write(pos_t pos, void *buf, size_t len);

	SharedFile &operator=(SharedFile &f);
	
	friend inline read(SharedFile &f, pos_t pos, void *buf, size_t len)
		{f.Read(pos, buf, len);};

 	friend inline write(SharedFile &f, pos_t pos, void *buf, size_t len)
 		{f.Write(pos, buf, len);};

	friend inline append(SharedFile &f, void *buf, size_t len)
		{f.Append(buf, len);};
};

inline clear(LockedFile &f, pos_t pos, size_t len)
	{return f.Clear(pos, len);};

inline append(LockedFile &f, void *buf, size_t len)
	{return f.Append(buf, len);};

inline request(LockedFile &f, pos_t pos, void *buf, size_t len)
	{return f.Request(pos, buf, len);};

inline update(LockedFile &f, pos_t pos, void *buf, size_t len)
	{return f.Update(pos, buf, len);};

#endif

