// Copyright (C) 1999 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of APE.
// 
// The exception is that, if you link the APE library with other files
// to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the APE library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name APE.  If you copy code from other releases into a copy of
// APE, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for APE, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "config.h"
#include "macros.h"
#include "thread.h"
#include "file.h"
#include "serial.h"
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>

static	bool	set_speed(long speed, termios *t)
{
	speed_t	rate = 0;

	switch(speed)
	{
	case 115200:
#ifdef	B115200
		rate = B115200;
		break;
#endif
	case 57600:
#ifdef	B57600
		rate = B57600;
		break;
#endif
	case 38400:
#ifdef	B38400
		rate = B38400;
		break;
#endif
	case 19200:
		rate = B19200;
		break;
	case 9600:
		rate = B9600;
		break;
	case 4800:
		rate = B4800;
		break;
	case 2400:
		rate = B2400;
		break;
	case 1200:
		rate = B1200;
		break;
	case 600:
		rate = B600;
		break;
	case 300:
		rate = B300;
		break;
	case 110:
		rate = B110;
		break;
	}
	if(!rate)
		return false;
	cfsetospeed(t, rate);
	cfsetispeed(t, rate);
	return true;
}

static	void	set_flow(int flow, termios *t)
{
	t->c_cflag &= ~CRTSCTS;
	t->c_iflag &= ~(IXON | IXANY | IXOFF);

	switch(flow)
	{
	case SERIAL_FLOW_SOFT:
		t->c_iflag |= (IXON | IXANY | IXOFF);
		break;
	case SERIAL_FLOW_BOTH:
		t->c_iflag |= (IXON | IXANY | IXOFF);
	case SERIAL_FLOW_HARD:
		t->c_cflag |= CRTSCTS;
		break;
	}
}

static	void set_format(int format, termios *t)
{
	unsigned parity = 0;
	unsigned csize = CS8;
	unsigned stop = 0;

	switch(format)
	{
	case SERIAL_FORMAT_7E1:
		parity = PARENB;
		csize = CS7;
		break;
	case SERIAL_FORMAT_7O1:
		parity = PARENB | PARODD;
		csize = CS7;
		break;
	case SERIAL_FORMAT_7N1:
		csize = CS7;
		break;
	case SERIAL_FORMAT_7S2:
		csize = CS7;
		stop = CSTOPB;
		break;
	case SERIAL_FORMAT_8E1:
		parity = PARENB;
		csize = CS8;
		break;
	case SERIAL_FORMAT_8O1:
		parity = PARENB;
		csize = CS8;
		break;
	case SERIAL_FORMAT_8N1:
		csize = CS8;
		break;
	case SERIAL_FORMAT_8S2:
		csize = CS8;
		stop = CSTOPB;
	}

	t->c_cflag &= ~CSIZE;
	t->c_cflag &= ~(PARENB | PARODD);
	t->c_cflag &= ~CSTOPB;
	t->c_cflag |= (stop | parity | csize) ;
}

static	void set_vtimer(int fd, termios *t, int min, int timer)
{
	if(min == t->c_cc[VMIN] && timer == t->c_cc[VTIME])
		return;

	t->c_cc[VMIN] = min;
	t->c_cc[VTIME] = timer;
	tcsetattr(fd, TCSANOW, t);
}

Serial::~Serial()
{
	if(!isatty(_fd))
		return;

	tcdrain(_fd);
	tcflush(_fd, TCIOFLUSH);
	tcsetattr(_fd, TCSANOW, &_restore);
	close(_fd);
	remove(_lockname);
}

#ifndef	O_NONBLOCK
#define	O_NONBLOCK	O_NDELAY
#endif

#ifndef	LOCK_FORMAT
#define	LOCK_FORMAT	"%010d %-64s"
#endif

Serial::Serial(const char *fname, long speed, int format, int flow)
{
	struct	stat ino;
	char	*cp;
	char	query[96];
	int	rc;
	unsigned pid;

	_nl = 13;

	if(stat("/var/lock", &ino))
		strcpy(_lockname, "/usr/spool/uucp/");
	else
		strcpy(_lockname, "/var/lock/");	

#ifdef	HAVE_REALPATH
	char path[129];
	realpath(fname, path);
	cp = strrchr(path, '/');
	if(cp)
		++cp;
	else
		cp = path;
#else
	cp = strrchr(fname, '/');
	if(cp)
		++cp;
	else
		cp = fname;
#endif
	strcat(_lockname, cp);
	
	for(;;)
	{
		_fd = open(_lockname, O_WRONLY | O_CREAT | O_EXCL, 0660);
		if(_fd > 0)
		{
			sprintf(query, LOCK_FORMAT, getpid(), "startup");
			write(_fd, query, strlen(query));
			close(_fd);
			break;
		}
		if(_fd < 0 && errno != EEXIST)
			throw(this);

		_fd = open(_lockname, O_RDONLY);
		if(_fd < 0)
		{
			if(errno == ENOENT)
				continue;

			throw(this);
		}
		ape_sleep(2000);
		rc = read(_fd, query, sizeof(query) - 1);
		if(rc < 1)
		{
			close(_fd);
			continue;
		}
		query[rc] = 0;
		pid = atoi(query);
		if(pid)
		{
			if(pid == getpid())
			{
				close(_fd);
				remove(_lockname);
				continue;
			}
	
			rc = kill(pid, 0);
			if(!rc || (errno == EPERM))
			{
				close(_fd);
				throw(this);
			}
		}
		close(_fd);
		remove(_lockname);
	}					 
		
	_fd = open(fname, O_RDWR | O_NONBLOCK | O_NOCTTY);
	if(_fd < 0)
	{
		remove(_lockname);
		throw(this);
	}

	tcgetattr(_fd, &_restore);
	tcgetattr(_fd, &_current);
	
	_current.c_oflag = 0;
	_current.c_lflag = 0;
	_current.c_iflag = 0;
	_current.c_cflag = CLOCAL | CREAD | HUPCL;
	memset(&_current.c_cc, 0, sizeof(_current.c_cc));
	_current.c_cc[VMIN] = 1;

	set_speed(speed, &_current);
	set_format(format, &_current);
	set_flow(flow, &_current);
	tcsetattr(_fd, TCSANOW, &_current);
	fcntl(_fd, F_SETFL, O_RDWR);
}

void Serial::setFormat(int format)
{
	set_format(format, &_current);
	tcsetattr(_fd, TCSANOW, &_current);
}

void Serial::setSpeed(long speed)
{
	set_speed(speed, &_current);
	tcsetattr(_fd, TCSANOW, &_current);
}

void Serial::setFlowControl(int flow)
{
	set_flow(flow, &_current);
	tcsetattr(_fd, TCSANOW, &_current);
}

int Serial::Inkey(timeout_t timeout)
{
	unsigned char buf;
	int	rc;

	timeout /= 100;
	set_vtimer(_fd, &_current, 0, timeout);
	
	rc = read(_fd, &buf, 1);
	if(rc < 1)
		return -1;

	return (int)buf;
}

int Serial::Read(void *buf, size_t len, timeout_t timeout)
{
	timeout /= 100;
	set_vtimer(_fd, &_current, len, timeout);

	return read(_fd, (char *)buf, len);
}

int Serial::Readline(char *buf, size_t max, timeout_t timeout)
{
	int key;
	int len = 0;

	while(len < max)
	{
		key = Inkey(timeout);
		if(key == -1)
			break;

		if(key == _nl)
		{
			++len;
			*(buf++) = '\n';
			break;
		}

		if(key == 10)
			continue;

		++len;
		*(buf++) = (char)key;
	}
	*buf = 0;
	return len;
}
