// Copyright (C) 1999 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of APE.
// 
// The exception is that, if you link the APE library with other files
// to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the APE library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name APE.  If you copy code from other releases into a copy of
// APE, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for APE, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "config.h"
#include "macros.h"
#include "thread.h"
#include "socket.h"
#include <fcntl.h>
#include <errno.h>

#ifndef	O_NONBLOCK
#define	O_NONBLOCK	O_NDELAY
#endif

static	void endsocket(int so)
{
#ifdef	SO_LINGER
	struct linger linger;

	linger.l_onoff = linger.l_linger = 0;
	setsockopt(so, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger));
#endif	
	shutdown(so, 2);
	close(so);
}

Socket::Socket(int domain, int type, int protocol)
{
	state = SOCKET_INITIAL;
	so = socket(domain, type, protocol);
	if(so < 0)
		throw(this);

	state = SOCKET_AVAILABLE;
}

Socket::Socket(int fd)
{
	state = SOCKET_INITIAL;
	if(fd < 0)
	{
		throw(this);
		return;
	}
	so = fd;
	state = SOCKET_AVAILABLE;
}

Socket::Socket(const Socket &orig)
{
	state = SOCKET_INITIAL;
	so = dup(orig.so);
	if(so < 0)
		throw(this);
	state = orig.state;
}

Socket::~Socket()
{
	if(state == SOCKET_INITIAL)
		return;

	endsocket(so);
	state = SOCKET_INITIAL;
	so = -1;
}		

Socket &Socket::operator=(const Socket &from)
{
	if(so == from.so)
		return *this;

	if(state != SOCKET_INITIAL)
		endsocket(so);

	so = dup(from.so);
	if(so < 0)
		state = SOCKET_INITIAL;
	else
		state = from.state;

	return *this;
}

void Socket::setCompletion(int mode)
{
	int flags = fcntl(so, F_GETFL);

	switch(mode)
	{
	case SOCKET_COMPLETION_IMMEDIATE:
		flags |= O_NONBLOCK;
		fcntl(so, F_SETFL, flags);
		break;
	case SOCKET_COMPLETION_DELAYED:
		flags &=~ O_NONBLOCK;
		fcntl(so, F_SETFL, flags);
		break;
	}
}

UDPSocket::UDPSocket(void) : 
Socket(AF_INET, SOCK_DGRAM, 0)
{
}

UDPSocket::UDPSocket(InetHostAddress &host, short port) :
Socket(AF_INET, SOCK_DGRAM, 0)
{
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_addr = getaddress(host);
	addr.sin_port = htons(port);
	if(connect(so, (sockaddr *)&addr, sizeof(addr)))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}
}		

UDPSocket::UDPSocket(InetAddress &ia, short port) : 
Socket(AF_INET, SOCK_DGRAM, 0)
{
	struct	sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_addr = getaddress(ia);
	addr.sin_port = htons(port);
	if(bind(so, (sockaddr *)&addr, sizeof(addr)))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}
	state = SOCKET_BOUND;
}

UDPSocket::UDPSocket(BroadcastAddress &ia, short port) : 
Socket(AF_INET, SOCK_DGRAM, 0)
{
	struct sockaddr_in addr;
	int	bcast = 1;

	addr.sin_family = AF_INET;
	addr.sin_addr = getaddress(ia);
	addr.sin_port = htons(port);

	if(bind(so, (sockaddr *)&addr, sizeof(addr)))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}

	if(setsockopt(so, SOL_SOCKET, SO_BROADCAST, &bcast, sizeof(bcast)))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}
	state = SOCKET_BOUND;
}

int UDPSocket::Write(InetHostAddress &host, short port, void *addr, size_t len)
{
	struct sockaddr_in to;

	to.sin_family = AF_INET;
	to.sin_addr = getaddress(host);
	to.sin_port = htons(port);	

	return sendto(so, addr, len, 0, (struct sockaddr *)&to, sizeof(to));
}

InetHostAddress UDPSocket::Peek(void)
{
	struct sockaddr_in from;
	InetHostAddress host;
	char buf;
	size_t len = sizeof(from);

	int rc = recvfrom(so, &buf, 1, MSG_PEEK, (struct sockaddr *)&from, &len);
	if(rc == 1)
		host = from.sin_addr;
	return host;
}

TCPSocket::TCPSocket(InetHostAddress &host, short port) : 
Socket(AF_INET, SOCK_STREAM, 0)
{
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_addr = getaddress(host);
	addr.sin_port = htons(port);

	if(connect(so, (struct sockaddr *)&addr, sizeof(addr)))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}
	state = SOCKET_CONNECTED;
}

TCPSocket::TCPSocket(InetAddress &ia, short port, int backlog) : 
Socket(AF_INET, SOCK_STREAM, 0)
{
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_addr = getaddress(ia);
	addr.sin_port = htons(port);

	if(bind(so, (struct sockaddr *)&addr, sizeof(addr)))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}

	if(listen(so, backlog))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}
	state = SOCKET_BOUND;
}

TCPSocket::TCPSocket(TCPSocket &server) : 
Socket(accept(server.so, NULL, NULL)) 
{
	if(!server.OnAccept(getPeer()))
	{
		endsocket(so);
		state = SOCKET_INITIAL;
		throw((Socket *)this);
		return;
	}
	state = SOCKET_CONNECTED;
}

TCPSocket::TCPSocket(void) :
Socket(AF_INET, SOCK_STREAM, 0) {};

InetHostAddress TCPSocket::getHost(void)
{
	InetHostAddress buf;
	struct sockaddr_in addr;
	size_t	len = sizeof(addr);

	if(getsockname(so, (struct sockaddr *)&addr, &len))
		return buf;

	buf = addr.sin_addr;
	return buf;
}

InetHostAddress TCPSocket::getPeer(void)
{
	InetHostAddress buf;
	struct sockaddr_in addr;
	size_t len = sizeof(addr);

	if(getpeername(so, (struct sockaddr *)&addr, &len))
		return buf;

	buf = addr.sin_addr;
	return buf;
}

int TCPSocket::Readline(char *buf, size_t max)
{
	int pos = 0;
	int len;
	int eol;
	bool nl = false;

	while(!nl)
	{
		len = Peek(buf + pos, max - pos);
		if(len <= 0)
		{
			if(!pos)
				return len;

			break;
		}
		eol = 0;
		while(eol <= len && !nl)
		{
			if(buf[pos + eol] == '\n')
				nl = true;

			++eol;
		}
		len = Read(buf + pos, eol);
		if(len < 0)
			break;
							 		
		if(len < eol)
			nl = false;

		pos += len;
	}
	buf[pos] = 0;
	return pos;
}

TCPSession::TCPSession(Semaphore *start, InetHostAddress &ia, short port, int pri, int stack) :
TCPSocket(), Thread(start, pri, stack)
{
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr = getaddress(ia);

	setCompletion(SOCKET_COMPLETION_DELAYED);	
	if(connect(so, (struct sockaddr *)&addr, sizeof(addr)))
	{
		if(errno == EINPROGRESS)
		{
			state = SOCKET_CONNECTING;
			return;
		}
		endsocket(so);
		state = SOCKET_INITIAL;
		return;
	}
	state = SOCKET_CONNECTED;
}
	

TCPSession::TCPSession(Semaphore *start, TCPSocket &s, int pri, int stack) :
TCPSocket(s), Thread(start, pri, stack)
{
	setCompletion(SOCKET_COMPLETION_IMMEDIATE);
}

int TCPSession::WaitConnection(timeout_t timeout)
{
	long flags = 0;
	fd_set out;
	struct timeval	timer;
	size_t len = sizeof(flags);

	switch(state)
	{
	case SOCKET_INITIAL:
		return -1;
	case SOCKET_CONNECTED:
		break;
	case SOCKET_CONNECTING:	
		timer.tv_sec = timeout / 1000;
		timer.tv_usec = (timeout % 1000) * 1000;
		FD_ZERO(&out);
		FD_SET(so, &out);
	
		if(select(so + 1, NULL, &out, NULL, &timer) != 1)
		{
			endsocket(so);
			state = SOCKET_INITIAL;
			return -1;
		}

		getsockopt(so, SOL_SOCKET, SO_ERROR, &flags, &len);
		if(flags)
		{
			endsocket(so);
			state = SOCKET_INITIAL;
			return -1;
		}
	}
	state = SOCKET_CONNECTED;
	return 0;	
}

void TCPSession::Initial(void)
{
	if(WaitConnection(60000))
		Exit();

	setCompletion(SOCKET_COMPLETION_IMMEDIATE);
}

