// Copyright (C) 1999 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of APE.
// 
// The exception is that, if you link the APE library with other files
// to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the APE library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name APE.  If you copy code from other releases into a copy of
// APE, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for APE, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "config.h"
#include "macros.h"
#include "thread.h"

typedef	void	*(*exec_t)(void *);
typedef	RETSIGTYPE (*signal_t)(int);

#ifdef	SIGCANCEL

static	RETSIGTYPE	sigcancel(int)
{
	pthread_exit(NULL);
}

#endif

static	RETSIGTYPE	sigsuspend(int)
{
	sigset_t sigs;

	sigemptyset(&sigs);
	sigaddset(&sigs, SIGUSR3);
	sigwait(&sigs, NULL);
}

static	sigset_t *blocked_signals(sigset_t *sig)
{
	sigemptyset(sig);
	sigaddset(sig, SIGINT);
	sigaddset(sig, SIGKILL);
	sigaddset(sig, SIGHUP);
	sigaddset(sig, SIGABRT);
	sigaddset(sig, SIGALRM);
	sigaddset(sig, SIGUSR3);
	return sig;
}

pthread_t Thread::_main = 0;
Thread	*Thread::_timer = NULL;
Mutex	Thread::_arm;

Thread::Thread(Semaphore *start, int pri, size_t stack)
{
	static	bool initial = false;
	struct	sigaction act;
	int	rc;

	if(!initial)
	{
		act.sa_handler = (signal_t)Timer;
		sigemptyset(&act.sa_mask);
#ifdef	SA_INTERRUPT
		act.sa_flags = SA_INTERRUPT;
#else
		act.sa_flags = 0;
#endif
		sigaction(SIGALRM, &act, NULL);

		act.sa_handler = sigsuspend;
		sigemptyset(&act.sa_mask);
#ifdef	SA_RESTART
		act.sa_flags = SA_RESTART;
#endif
		sigaction(SIGUSR3, &act, NULL);

#ifdef	SIGCANCEL
		act.sa_flags = 0;
		act.sa_handler = sigcancel;
		sigemptyset(&act.sa_mask);
		sigaddset(&act.sa_mask, SIGALRM);

		sigaction(SIGCANCEL, &act, NULL);
#endif
		_main = pthread_self();
		First();
		initial = true;
	}

	pthread_attr_init(&_attr);
	pthread_attr_setdetachstate(&_attr, PTHREAD_CREATE_JOINABLE);
	pthread_attr_setinheritsched(&_attr, PTHREAD_INHERIT_SCHED);

#ifdef	PTHREAD_STACK_MIN
	if(pthread_attr_setstacksize(&_attr,
		stack <= PTHREAD_STACK_MIN ? PTHREAD_STACK_MIN : stack))
			throw(this);
#endif

#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING	
	if(pri)
	{
		struct sched_param sched;
		int policy;

		policy = sched_getscheduler(0);
		if(policy < 0)
			throw(this);

		sched_getparam(0, &sched);

		pri = sched.sched_priority - pri;
		if(pri 	> sched_get_priority_max(policy))
			pri = sched_get_priority_max(policy);

		if(pri < sched_get_priority_min(policy))
			pri = sched_get_priority_min(policy);

		sched.sched_priority = pri;
		pthread_attr_setschedparam(&_attr, &sched);
	}	
#endif
	_parent = pthread_self();
	_cancel = THREAD_CANCEL_INITIAL;
	_start = start;
	_tid = 0;

	if(_start)
	{
		rc = pthread_create(&_tid, &_attr, exec_t(&Execute), this);
		if(rc)
			throw(this);
	}
};

Thread::Thread(const Thread &th)
{
	sigset_t mask, newmask;
	int rc;

	if(th._tid == th._parent)
	{
		throw(this);
		return;
	}
	
	_parent = pthread_self();
	_cancel = THREAD_CANCEL_INITIAL;
	pthread_sigmask(SIG_BLOCK, blocked_signals(&newmask), &mask);
	rc = pthread_create(&_tid, &_attr, exec_t(&Execute), this);
	pthread_sigmask(SIG_SETMASK, &mask, NULL);
	if(rc)
		throw(this);
}

int Thread::Start(Semaphore *start)
{
	if(_tid)
		return -1;

	_start = start;
	return pthread_create(&_tid, &_attr, exec_t(&Execute), this);
}

void Thread::Terminate(void)
{
	if(!_tid)
		return;

	if(pthread_self() != _tid)
	{
		pthread_cancel(_tid);
		pthread_join(_tid, NULL);
	}
	_tid = pthread_self();
}

void Thread::Timer(int signo)
{
	if(_timer)
	{
		_timer->_alarm = 0;
		_timer->OnTimer();
	}
}

void	Thread::Execute(Thread *th)
{
	sigset_t mask;
	pthread_t tid;

	pthread_sigmask(SIG_BLOCK, blocked_signals(&mask), NULL);
	th->setCancel(THREAD_CANCEL_INITIAL);
	th->Yield();
	if(th->_start)
	{
		th->_start->Wait();
		th->_start = NULL;
	}

	th->Initial();
	if(th->getCancel() == THREAD_CANCEL_INITIAL)
		th->setCancel(THREAD_CANCEL_DEFAULT);

	if(!setjmp(th->_env))
		th->Run();

	if(th == _timer)
		--_arm;

	th->setCancel(THREAD_CANCEL_DISABLED);
	th->Final();
	pthread_detach(th->_tid);
	pthread_exit(NULL);
}

void	Thread::setTimer(timeout_t timer)
{
	sigset_t sigs;

	timer /= 1000;
	++_arm;
	_timer = this;
	time(&_alarm);
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGALRM);
	pthread_sigmask(SIG_UNBLOCK, &sigs, NULL);
	alarm(timer);
}	

timeout_t Thread::getTimer(void)
{
	time_t now;
	if(!_alarm)
		return 0;

	time(&now);
	return (timeout_t)(((now - _alarm) * 1000) + 500);
}

void Thread::endTimer(void)
{
	sigset_t sigs;
	if(_timer != this)
		return;

	alarm(0);
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGALRM);
	pthread_sigmask(SIG_BLOCK, &sigs, NULL);
	--_arm;
	_timer = NULL;
}

void	Thread::WaitSignal(signo_t signo)
{
	sigset_t	mask;

	sigemptyset(&mask);
	sigaddset(&mask, signo);
	sigwait(&mask, &signo);
}	

void	Thread::setSuspend(int mode)
{
	sigset_t mask;

	sigemptyset(&mask);
	sigaddset(&mask, SIGUSR3);

	switch(mode)
	{
	case THREAD_SUSPEND_ENABLE:
		pthread_sigmask(SIG_UNBLOCK, &mask, NULL);
		return;
	case THREAD_SUSPEND_DISABLE:
		pthread_sigmask(SIG_BLOCK, &mask, NULL);
	}
}

#ifdef	SIGCANCEL

void	Thread::setCancel(int mode)
{
	sigset_t	mask;

	sigemptyset(&mask);
	sigaddset(&mask, SIGCANCEL);
	
	switch(mode)
	{
	case THREAD_CANCEL_IMMEDIATE:
		pthread_sigmask(SIG_UNBLOCK, &mask, NULL);
		break;
	case THREAD_CANCEL_INITIAL:
	case THREAD_CANCEL_DISABLED:
	case THREAD_CANCEL_DEFERRED:
		pthread_sigmask(SIG_BLOCK, &mask, NULL);
		break;
	default:
		mode = -1;
	}
	if(mode >-1)
		_cancel = mode;
}
#else

void	Thread::setCancel(int mode)
{
	switch(mode)
	{
	case THREAD_CANCEL_IMMEDIATE:
		pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
		break;
	case THREAD_CANCEL_DEFERRED:
		pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
		break;
	case THREAD_CANCEL_INITIAL:
	case THREAD_CANCEL_DISABLED:
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
		break;
	default:
		mode = -1;
	}
	if(mode > -1)
		_cancel = mode;
}

#endif

void	Thread::Yield(void)
{
#ifdef	SIGCANCEL
	sigset_t cancel, old;

	sigemptyset(&cancel);
	sigaddset(&cancel, SIGCANCEL);

	if(_cancel != THREAD_CANCEL_DISABLED && 
	   _cancel != THREAD_CANCEL_INITIAL)
		pthread_sigmask(SIG_UNBLOCK, &cancel, &old);
#else
	pthread_testcancel();
#endif
#ifdef	HAVE_PTHREAD_YIELD
	pthread_yield();
#endif

#ifdef	SIGCANCEL
	if(_cancel != THREAD_CANCEL_DISABLED)
		pthread_sigmask(SIG_SETMASK, &old, NULL);
#endif
}

