// Copyright (C) 1999 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of APE.
// 
// The exception is that, if you link the APE library with other files
// to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the APE library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name APE.  If you copy code from other releases into a copy of
// APE, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for APE, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#ifndef	__APE_THREAD_H__
#define	__APE_THREAD_H__
#define	__APE_POSIX

#ifndef	HAVE_PTHREAD_H
#include <pthread.h>
#include <semaphore.h>
#endif

#ifndef	__APE_MACROS_H__
#include <APE/macros.h>
#endif

#include <time.h>
#include <signal.h>
#include <setjmp.h>

typedef	pthread_t	tid_t;
typedef	unsigned long	timeout_t;
typedef	int		signo_t;

#define	ENTER_CRITICAL	EnterMutex();
#define	LEAVE_CRITICAL	LeaveMutex();
#define	ENTER_DEFERRED	setCancel(THREAD_CANCEL_DEFERRED);
#define LEAVE_DEFERRED 	setCancel(THREAD_CANCEL_IMMEDIATE);

// These macros override common functions with thread-safe versions. In
// particular the common "libc" sleep() has problems since it normally
// uses SIGARLM (as actually defined by "posix").  The pthread_delay and
// usleep found in libpthread are gaurenteed not to use SIGALRM and offer
// higher resolution.  psleep() is defined to call the old process sleep.

#define	sleep(x)	ape_sleep((x) * 1000)
#define	yield()		ape_yield()
#define	psleep(x)	(sleep)(x)

enum
{
	THREAD_CANCEL_INITIAL=0,
	THREAD_CANCEL_DEFERRED=1,
	THREAD_CANCEL_IMMEDIATE,
	THREAD_CANCEL_DISABLED,
	THREAD_CANCEL_DEFAULT=THREAD_CANCEL_DEFERRED
};

#ifndef	SIGUSR3
#ifdef	SIGWINCH
#define	SIGUSR3	SIGWINCH
#else
#define	SIGUSR3	SIGINT
#endif
#endif

class Mutex
{
protected:
	pthread_mutex_t	_mutex;

public:
	Mutex();

	~Mutex()
		{pthread_mutex_destroy(&_mutex);};

	inline void EnterMutex(void)
		{pthread_mutex_lock(&_mutex);};

	inline void LeaveMutex(void)
		{pthread_mutex_unlock(&_mutex);};
};

class	MutexCounter : public Mutex
{
private:
	int	counter;

public:
	MutexCounter();
	
	friend int operator ++(MutexCounter &mc);
	friend int operator --(MutexCounter &mc);
};

class Semaphore
{
protected:
	sem_t _semaphore;

public:
	Semaphore(size_t resource = 0);

	~Semaphore()
		{sem_destroy(&_semaphore);};

	inline void Wait(void)
		{sem_wait(&_semaphore);};
	
	inline void Post(void)
		{sem_post(&_semaphore);};
};

class Event : public Mutex
{
protected:
	pthread_cond_t _cond;
	bool _signaled;
	int _count;

public:
	Event();

	~Event()
		{pthread_cond_destroy(&_cond);};
	
	void Reset(void)
		{_signaled = false;};

	void Signal(void);
	void Wait(timeout_t timer = 0);
};

class Buffer
{
private:
	Mutex lock_head, lock_tail;
	Semaphore size_head, size_tail;
	size_t _size;
	size_t _used;

protected:
	virtual int OnPeek(void *buf) = 0;
	virtual int OnWait(void *buf) = 0;
	virtual int OnPost(void *buf) = 0;

public:
	Buffer(size_t capacity);
	virtual ~Buffer()
		{return;};

	inline size_t getSize(void)
		{return _size;};

	inline size_t getUsed(void)
		{return _used;};

	int Wait(void *buf);
	int Post(void *buf);
	int Peek(void *buf);
};

class FixedBuffer : public Buffer
{
private:
	char *buf, *head, *tail;
	size_t objsize;

protected:
	int OnPeek(void *buf);
	int OnWait(void *buf);
	int OnPost(void *buf);	

public:
	FixedBuffer(size_t capacity, size_t objsize);
	FixedBuffer(const FixedBuffer &fb);
	~FixedBuffer();

	FixedBuffer &operator=(const FixedBuffer &fb);
};

class Pipe
{
private:
	int fd[2];

protected:	// fork support
	inline void Sender(void)
		{close(fd[0]);};

	inline void Receiver(void)
		{close(fd[1]);};

public:
	Pipe();
	~Pipe();
	Pipe(const Pipe &orig);

	Pipe &operator=(const Pipe &orig);
	
	inline int Read(void *addr, size_t len)
		{return read(fd[0], (char *)addr, len);};

	inline int Write(void *addr, size_t len)
		{return write(fd[1], (char *)addr, len);};

	friend inline read(Pipe &p, void *addr, size_t len)
		{return read(p.fd[0], (char *)addr, len);};

	friend inline write(Pipe &p, void *addr, size_t len)
		{return write(p.fd[1], (char *)addr, len);};
};

class Thread
{
private:
	pthread_t _tid;
	pthread_t _parent;
	pthread_attr_t _attr;
	int	_cancel;
	jmp_buf	_env;
	time_t	_alarm;
	Semaphore *_start;
	
	static	pthread_t _main;
	static	Thread	*_timer;
	static	Mutex	_arm;

	static	void Execute(Thread *th);
	static	void Timer(int signo);

protected:
	enum
	{
		THREAD_SUSPEND_ENABLE,
		THREAD_SUSPEND_DISABLE
	};

	enum
	{
		THREAD_SIGNAL_BLOCKED=1,
		THREAD_SIGNAL_UNBLOCK,
		THREAD_SIGNAL_INSTALL,
		THREAD_SIGNAL_INTERRUPT,
		THREAD_SIGNAL_ONESHOT
	};

	virtual void Run(void) = 0;

	virtual void First(void)
		{return;};

	virtual void Final(void)
		{return;};

	virtual void Initial(void)
		{return;};

	inline void SignalParent(signo_t signo)
		{pthread_kill(_parent, signo);};

	inline void SignalMain(signo_t signo)
		{pthread_kill(_main, signo);};

	virtual void OnTimer(void)
		{return;};

	inline void Sleep(timeout_t msec)
			{ape_sleep(msec);};

	inline void Exit(void)
		{longjmp(_env, 1);};

	void setTimer(timeout_t timer);
	timeout_t getTimer(void);
	void endTimer(void);
	void WaitSignal(signo_t signo);
	void Yield(void);
	void setCancel(int mode);
	void setSuspend(int mode);
	void Terminate(void);

public:
	Thread(Semaphore *start = NULL, int pri = 0, size_t stack = 0);
	Thread(const Thread &th);
	virtual ~Thread()
		{Terminate();};

	int Start(Semaphore *start = NULL);

	inline void SignalThread(int signo)
		{pthread_kill(_tid, signo);};

	inline void Suspend(void)
		{pthread_kill(_tid, SIGUSR3);};

	inline void Resume(void)
		{pthread_kill(_tid, SIGUSR3);};

	inline int getCancel(void)
		{return _cancel;};

	inline bool isRunning(void)
		{return _tid != 0;};

	inline bool isThread(void)
		{return _tid == pthread_self();};

	friend void ape_sleep(timeout_t msec);

	friend void suspend(Thread &th)
		{pthread_kill(th._tid, SIGUSR3);};

	friend void resume(Thread &th)
		{pthread_kill(th._tid, SIGUSR3);};

	friend inline void operator++(Thread &th)
		{th._start->Post();};

	friend inline int start(Thread &th, Semaphore *start)
		{th.Start(start);};
};

inline void operator ++(Mutex &m)
	{m.EnterMutex();};
		
inline void operator --(Mutex &m)
	{m.LeaveMutex();};

inline void operator ++(Semaphore &s)
	{s.Post();};

inline void operator --(Semaphore &s)
	{s.Wait();};

inline void operator ++(Event &s)
	{s.Signal();};

inline void operator --(Event &s)
	{s.Wait();};

inline void signal(Thread &th, int signo)
	{th.SignalThread(signo);};

inline void signal(Event &ev)
	{ev.Signal();};

inline void signal(Semaphore &sem)
	{sem.Post();};

inline void wait(Semaphore &sem)
	{sem.Wait();};

inline void wait(Event &ev, timeout_t timer)
	{ev.Wait(timer);};

inline void reset(Event &ev)
	{ev.Reset();};

inline int get(Buffer &b, void *o)
	{return b.Wait(o);};

inline int put(Buffer &b, void *o)
	{return b.Post(o);};

inline int peek(Buffer &b, void *o)
	{return b.Peek(o);};

int operator++(MutexCounter &mc);
int operator--(MutexCounter &mc);

struct	timespec *gettimeout(struct timespec *spec, timeout_t timeout);	
void	ape_sleep(timeout_t msec);
void	ape_yield(void);
void	wait(signo_t signo);
void	pdetach(void);
#endif

